"""Terminal utilities."""
from mode.utils.compat import isatty
from .spinners import Spinner, SpinnerHandler
from .tables import Table, TableDataT, logtable, table

__all__ = [
    'Spinner',
    'SpinnerHandler',
    'Table',
    'TableDataT',
    'isatty',
    'logtable',
    'table',
]
