# Generated by Django 2.1.4 on 2018-12-24 12:13

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True,
                                        serialize=False,
                                        verbose_name='ID')),
                ('name', models.CharField(max_length=100,
                                          verbose_name='name')),
                ('score', models.DecimalField(decimal_places=1000,
                                              default=0.0,
                                              max_digits=1000,
                                              verbose_name='score')),
                ('active', models.BooleanField(default=True,
                                               verbose_name='active')),
            ],
            options={
                'verbose_name': 'account',
                'verbose_name_plural': 'accounts',
            },
        ),
    ]
