"""Faust Web."""
from .base import Request, Response, Web
from .blueprints import Blueprint
from .views import View, gives_model, takes_model
from faust.types.web import ResourceOptions

__all__ = [
    'Request',
    'Response',
    'ResourceOptions',
    'Web',
    'Blueprint',
    'View',
    'gives_model',
    'takes_model',
]
